$! Name:    linkshrlib.com
$! Version: 1.1.4
$! Purpose: Link an object library into a shared library, 
$!          using two C programs, if not available in 
$!          symbol listlib and vectorcompatible will
$!          be compiled.
$! Author:  Jose Baars, 2010
$!---------------------------------------------------------- 
$! 1.1.1    Listlib is able to process Unix version scripts
$! 1.1.2    Interpretmap can read both stream_lf and variable length         
$! 1.1.3    Added compile option
$! 1.1.4    Handle perverse line continuation in vectorcompatible
$! 1.1.5    Interpretmap init pointer to NULL
$! 1.1.6    Listlib missed some vectors due to strncmp; default -f l
$!---------------------------------------------------------- 
$ on error then goto Error
$!
$ if f$edit(p1,"trim,compress,upcase") .eqs. "COMPILE"
$ then
$	if f$type( listlib ) .nes. ""
$	then
$		set noon
$		delete/sym/glob listlib
$		delete/sym/glob vectorcompatible
$		if f$type( listlib ) .nes. ""
$		then
$			delete/sym listlib
$			delete/sym vectorcompatible
$		endif
$	endif
$  call MakeVectorcompatible
$  call MakeListLib
$	exit 
$ endif
$!
$ gosub Init
$!
$ call ListLib
$!
$ if oldvecname .nes. ""
$ then
$  call MakeVectorcompatible
$  call Vectorcompatible
$  vcstat  = $status
$  if vcstat .eq. 1
$  then
$		say "Vectorcompatible produced identical vector options, deleting ''f$parse(newvecname,,,"name")'.opt;*"
$		delete 'newvecname'*
$		newvecname = oldvecname
$  endif
$ endif
$!
$ call LinkObjectLibrary
$!
$exit
$Error:
$ say "LINKSHRLIB.COM ended in error" 
$ exit '$status'

$Init:
$! parse and check parameters
$!
$ say    = "write sys$output"
$ thisid = %x'f$getjpi(0,"pid")
$ arch   =  f$getsyi("arch_name")
$!
$! set up global symbols vectorcompatible
$! and listlib to ready made executables.
$!
$ if f$search("sys$startupo:linkshrlib_startup.com") .nes. ""
$ then
$		@sys$startup:linkshrlib_startup.com
$ endif
$!
$ objlibname = f$search(p1)
$ shrlibname = f$parse(p2) 
$!
$ validnames  = "help,demangle_db,exclude_symbols,old_vector,new_vector,link_options,format,version,ident,sharedlib,vmap,vlab" 
$ symbolnames = "xhelp,xdemangledb,xexcludesym,xoldvecname,xnewvecname,xlinkeropt,xvectorform,xlibversion,xident,xsharedlib,xvmap,xlab" 
$ par = p3 + " " + p4 + " " + p5 + " " + p6 + " " + p7 + " " + p8
$!
$ gosub ParseQual
$!
$ newvecname = f$edit(xnewvecname,"compress,trim")
$ demangledb = f$search(xdemangledb)
$ excludesym = xexcludesym
$ oldvecname = xoldvecname
$ linkeropt  = xlinkeropt
$ vectorform = xvectorform
$ libversion = xlibversion
$ libident   = xident
$ sharedlib  = xsharedlib
$ help		 = xhelp
$ vmap 		 = f$search(xvmap)
$ vlab 		 = xlab
$!
$ if p1 .eqs. "" .or. p2 .eqs. "" .or. help .nes. ""
$ then
$     goto Printhelp
$ endif
$!
$ if f$edit(p1,"upcase") .eqs. "HELP"
$ then
$     goto Printhelp
$ endif 
$!
$ if objlibname .eqs. ""
$ then 
$     say "Can't find objectlib."
$		exit %X18292
$ endif
$ if newvecname .eqs. ""
$ then
$     newvecname = f$parse( objlibname,,,"name") + "_vector.opt"
$     newvecname = f$parse( newvecname )
$ else
$     newvecname = f$parse(newvecname)
$ endif
$ if libversion .nes. ""
$ then
$        ymajorversion = f$element(0,".",libversion)
$        yminorversion = f$element(1,".",libversion)
$!				strip leading zeroes
$			Symajloop:
$			if f$extract(0,1,ymajorversion) .eqs. "0"
$			then
$				ymajorversion = ymajorversion - "0"
$				goto Symajloop
$			endif
$		   majorversion = ymajorversion
$			if ymajorversion .eqs. "" then majorversion = 0
$			Syminloop:
$			if f$extract(0,1,yminorversion) .eqs. "0"
$			then
$				yminorversion = yminorversion - "0"
$				goto Syminloop
$			endif
$		   minorversion = yminorversion
$			if yminorversion .eqs. "" then minorversion = 0
$!
$        if 'minorversion' .ne. f$integer( minorversion) .or. -
            'majorversion' .ne. f$integer( majorversion)
$        then
$           say "libversion should be format x.y, e.g. 1.23, or 1.010203"
$           goto PrintHelp
$        endif
$ else
$       if f$locate("/opt",f$edit(linkeropt,"lowercase")) .ge. f$length(linkeropt)
$       then
$           say " ATTENTION! will generate shared image with version 1.0!"
$           say " this may not be what you intend. Use an option file or "
$           say " use +version=x.y"
$           libversion  = "1.0"
$           majorversion = "1"
$           minorversion = "0"
$       endif     
$ endif
$!
$ if demangledb .eqs. "" .and. xdemangledb .nes. ""
$ then
$     say "Can't find demangle db"
$     goto Printhelp
$ endif
$!
$ if vmap .eqs. "" .and. xvmap .nes. ""
$ then
$     say "Can't find Unix version script ''xvmap'"
$     goto Printhelp
$ endif
$!
$ if vlab .nes. "" .and. vmap .nes. ""
$ then 
$		define/user sys$error  nl:
$		define/user sys$output nl:
$		search/exact 'vmap' "''vlab'"/noout
$		if $severity .ne. 1
$		then
$			say "Unix version label ''vlab' not found in version script."
$			goto PrintHelp
$		endif
$ endif
$! no check of the exclude string
$ if oldvecname .nes. ""
$ then
$     oldvecname = f$search( oldvecname )
$     if oldvecname .eqs. ""
$     then
$        say "Can't find old vector options file"
$        goto Printhelp
$     endif
$     if f$edit(f$element(0,";",newvecname),"upcase") .eqs. -
         f$edit(f$element(0,";",oldvecname),"upcase")
$     then
$        newvecname = f$parse(newvecname,,,"device") -
                    + f$parse(newvecname,,,"directory") -
                    + f$parse(newvecname,,,"name") + "_new" -
                    + f$parse(newvecname,,,"type")
$     endif
$ endif
$!
$ say "New linker vector options file will be called ''newvecname'"
$!
$! no check of p7, the other linker options
$ if xvectorform .nes. ""
$ then
$     vectorform = ""
$     p8f = f$edit(xvectorform,"lowercase,trim")
$     if p8f .eqs. "p" then vectorform = "-f p"
$     if p8f .eqs. "l" then vectorform = "-f l"
$     if vectorform .eqs. ""
$     then
$        vectorform = "-f p"
$        say "Unrecognized vector file format, will use p"
$     endif
$ endif
$!
$return
$!
$ParseQual:
$! say "par = [''par']"
$! initialize all qualifier symbols to ""
$ xcount = 0
$XLoop:
$ thissymbol = f$element(xcount,",",symbolnames)
$ if thissymbol .eqs. "," then goto EndXLoop
$  'thissymbol' = ""
$ xcount = xcount + 1
$ goto XLoop
$EndXLoop:
$ delete/symbol xcount
$! validate qualifiers and fill all qualifiers with values
$! first qualifier after + so start with 1
$qcount = 1
$ QLoop: 
$   thispar = f$element(qcount,"+",par)
$   if thispar .eqs. "+" then goto EndQLoop
$!
$   thisname  = f$edit(f$element(0,"=",thispar),"lowercase,trim")
$   thisvalue = f$edit(f$element(1,"=",thispar),"trim")
$   vcount = 0 
$   Vloop:
$     thisvalid = f$element(vcount,",",validnames)
$     if thisvalid .eqs. "," 
$     then 
$        if thisname .nes. ""
$        then
$           say "Unrecognized qualifier \''thisname'\"
$           say "Valid qualifiers : ''validnames'"
$        endif
$        goto EndVLoop
$     endif
$     thisnamelen = f$length(thisname)
$     if f$extract(0,thisnamelen,thisname) .eqs. f$extract(0,thisnamelen,thisvalid)
$     then 
$        thissymbol   = f$element(vcount,",",symbolnames)
$        if f$extract(0,1,thisvalue) .eqs. """" 
$        then 
$           thisvalue    = thisvalue - """" - """"
$        endif
$        'thissymbol' = thisvalue
$        goto EndVloop
$     endif
$     vcount = vcount + 1
$     goto VLoop
$  EndVLoop:
$!
$ qcount = qcount + 1
$ goto  QLoop
$EndQLoop:
$return
$!
$Printhelp:
$ type sys$input:

Create a shareable image from an objectlibrary

Uses 2 executables, that will be compiled and linked if 
not available in foreign symbols listlib and 
vectorcompatible.

Usage: @Linkshrlib help
       shows this screen.

Usage: @Linkshrlib <objectlib> <shrlibname> [[+qual=value],...]

       Qualifiers are preceded by + in stead of /, to not confuse 
       DCL. Make sure to omit spaces, to not surpass the 8 parameter
       limit of DCL. The first 3 characters of the qualifier 
       will be sufficient.
   
       +new_vector=<vector options filename to be generated>
                    If not supplied this will be 
                    <shrlibname>_vector.opt
                    If this produces the same filename as t
                    the old vectoroptions this name will be                              
                    <objlibname>_vector_new.opt
       +old_vector=<old vector options filename> 
                   New vectorfile is made upward compatible 
                   with this options file                    
       +format=[p|l] format of new vector options file
                   format can be p or l, default is l, to prevent
                   the linker running out of space:
                   p: symbol_vector=(a=PROCEDURE,b=PROCEDURE,...)
                   l: symbol_vector=(a=PROCEDURE)
                      symbol_vector=(b=PROCEDURE)
                      ...
       +demangle_db=<demangledb filename> 
                   Used if symbols are shortened by the C 
                   compiler to egenrate both upper and 
                   lowercase vectors or aliases.
                   Works not good if sources are not compiled
                   /names=(as_is)
       +exclude_symbols="sym1,sym2,*wildcard*"
                   Do not make vectors for possibly wildcarded
                   symbols in this comma separated list.
                   Commonly, symbols starting with _ (underscore)
                   are not to be made global. In this case, this
                   qualifier should be +exclude_symbols="_*".
      +vmap=<Unix version script filename>
                   Some packages provid a list of symbols to be
                   exported or not, per version, for OS's that
                   support versioning. This is a finer grained
                   method thatn the crude exclude_symbols list.
      +vlab="version name as used in Unix version script"
                   By default, the version with most dependencies
                   (usually the highest version) in the version
                   script will be used. But, you can specify it 
                   here. The label should be exactly the same as 
                   used in the version script.
      +link_options = "..."
                   Literal extra options that will be appended
                   to the linker commandline. Literal means that
                   to add an option file the preceding comma needs
                   to be included +link=",blurb.opt/opt"

                   Not having a GSMATCH statement as option when linking
                   a shared library will have images linked against 
                   it not want to accept newer versions.
                   Not having an IDENTIFICATION is very poor practice.
                   If no option file could be detected, and no +version
                   qualifier is present, version 1.0 is used.
      +version = x.y    x and y must be decimal numbers
                   If you do not want to write an optionfile with GSMATCH 
                   and IDENTIFICATION statements, you can give libversion
                   on the command line, and an options file 
                   called <objlibname>_version.opt is created with 
                   GSMATCH=LEQUAL,x,y              
                   IDENTIFICATION="<shrlibname> x.y"
                   that will be apppended to the linker commandline
                   before the link_options if supplied.           
      +ident       see version. string replaceing the automatically
                   generated ID when only version is supplied.

      +sharedlib   if the shared library needs to be linked against
                   other shared libraries, give a comma separated list 
                   of logicals or names of the shared libraries. 
                   This is an alternative to stating them in a separate 
                   optionfile. These options will be written in 
                   <objlibname>_version.opt even if version is not 
                   specified.
            
   Example:
   @linkshrlib one2.olb  one2shr.exe +version=1.0
      will create one2shr.exe, one2shr_vector.opt and one2shr_version.opt

   @linkshrlib one21.olb one2shr.exe +version=1.1 +old=one2_vector.opt   
      will create new version of one2shr.exe, one2shr_vector_new.opt and
      one2shr_version.opt. If binary compatible, this version will be 
      upward compatible with the previous shareable image.

$exit
$!
$!--------------------------------------------------------------------
$!
$Listlib: subroutine
$!
$! create a new vector file, if demangele db is given, with aliases
$!
$ on error then goto ListlibError
$!
$ call MakeListLib
$!
$ library/extract=*/out=linkshrlib_objects_tmp.'thisid' 'objlibname'
$!
$ if arch .eqs. "IA64"
$ then
$  analyzequal = "/section=symtab"
$ endif
$!
$ if arch .eqs. "Alpha"
$ then
$  analyzequal = "/gsd"
$ endif
$!
$ define/user sys$error  nl:
$ define/user sys$output nl:
$ analyze/object'analyzequal' linkshrlib_objects_tmp.'thisid' - 
         /out=linkshrlib_objanalysis_tmp.'thisid'
$!
$ if .not. $status then goto ListlibError
$!
$ minusd = ""
$ if demangledb .nes. "" then minusd = "-d ""''demangledb'"""
$!
$ minuse = ""
$ if excludesym .nes. "" then minuse  = "-e ""''excludesym'"""
$!
$ minusm = ""
$ if vmap .nes. "" then minusm = "-m ""''vmap'"""
$!
$ minusn = ""
$ if vlab .nes. "" then minusn = "-n ""''vlab'"""
$!
$ listlib "''objlibname'" -
         -f l -
	      -a linkshrlib_objanalysis_tmp.'thisid' -
         -o "''newvecname'" 'minusd' 'vectorform' -
			'minusm' 'minusn' 'minuse'
$!
$ if f$search("linkshrlib_*_tmp.''thisid';*") .nes. ""
$ then
$  delete linkshrlib_*_tmp.'thisid';*
$ endif
$!
$exit
$!
$ListlibError:
$!
$ stat = $status
$ say "Error in listlib"
$ if f$search("linkshrlib_*_tmp.''thisid';*") .nes. ""
$ then
$  delete linkshrlib_*_tmp.'thisid';*
$ endif
$ exit 'stat'
$endsubroutine
$!
$!--------------------------------------------------------------------
$!
$VectorCompatible: subroutine
$! generate an upward compatible vectorfile.
$ set noon 
$ vectorcompatible -f l -s 'oldvecname' -n 'newvecname' -
                   -o 'newvecname' 'vectorform'
$ vcstat = $status
$ exit 'vcstat'
$endsubroutine  
$!
$!--------------------------------------------------------------------
$!
$LinkObjectLibrary: subroutine
$!
$ say "Link ''objlibname' to ''shrlibname'"
$ on error then exit
$ versionopt = ""
$ if libversion .nes. "" .or. sharedlib .nes. ""
$ then
$     shrname = f$parse(shrlibname,,,"name")
$     if libident .nes. ""
$     then
$        shrname = libident
$ 		else
$			shrname = "''shrname' V''majorversion'.''minorversion'"
$     endif 
$!		let linker choke and give a message if the id is too long
$!
$     libversionfile = f$parse(objlibname,,,"name") + "_version.opt"
$     open/write libv 'libversionfile'
$     if libversion .nes. ""
$     then
$      write libv "GSMATCH=LEQUAL,''majorversion',''minorversion'"
$      if libident .nes. "" then write libv "CASE_SENSITIVE=YES"
$      write libv "IDENTIFICATION=""''shrname'"""
$      if libident .nes. "" then write libv "CASE_SENSITIVE=NO"
$     endif
$     if sharedlib .nes. ""
$     then
$        scount = 0
$        ShLoop:
$           thislib = f$element(scount,",",sharedlib)
$           if thislib .eqs. "," then goto EndShLoop
$           write libv "''thislib'/share"
$           scount = scount + 1
$        goto ShLoop
$        EndShLoop:
$     endif
$     close libv
$     say "Using version option file ''f$parse(libversionfile)'"
$     say "Using vector  option file ''f$parse(newvecname)'"
$!
$     versionopt = "," + f$parse(libversionfile) + "/opt"
$ endif
$!
$ link/share/exe='shrlibname' 'objlibname'/lib ,'newvecname'/opt 'versionopt' 'linkeropt' 
$!
$exit
$endsubroutine
$!
$!--------------------------------------------------------------------
$! C-code and build of vectorcompatible and listlib 
$! underneath. Snippable if executables are available.
$!--------------------------------------------------------------------
$!
$MakeVectorCompatible: subroutine
$ set noon
$ if f$type( vectorcompatible ) .nes. ""
$ then
$      define/user sys$output nl:
$      define/user sys$error  nl:
$		 vectorcompatible
$      if f$message($status,"ident") .eqs. "%BADPARAM" then exit 1     
$ endif  
$!
$ say "Compile and link vectorcompatible"
$ type/out=vectorcompatible.c sys$input:
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stat.h>

#include <errno.h>
#include <errnodef.h>
#include <ssdef.h>

/* assumptions: vectors are defined with
   aliases like this
symbol_vector=( -
Compress=PROCEDURE, -
compress/Compress=PROCEDURE, -
COMPRESS/Compress=PROCEDURE )

or like this

symbol_vector=(Compress=PROCEDURE)
symbol_vector=(compress/Compress=PROCEDURE)
symbol_vector=(COMPRESS/Compress=PROCEDURE)

Jose Baars 2010.
*/

#define VALIAS  	1
#define VVECTOR   2
#define VCOMMENT  3

typedef struct image_vector{
char	*name;   
char	*vectorname;  	/* if alias this is the real vector name */
char	*dtype;        /* vector to what: data, procedure or private_procedure */
int	vtype;  			/* alias: VALIAS or symbol vector: VVECTOR*/
int	id;           	/* from what file do you come from */
int	written;			
struct image_vector *next;
} ivector, *ivectorPtr;

/*------------------------------------------------------------------*/
char *vtype2string(int vtype )
{
char	*vtype_names[]={"undefined","ALIAS","VECTOR","COMMENT"};

if ( vtype > 3 )return( "Illegal");
return( vtype_names[ vtype ] );
}

/*------------------------------------------------------------------*/
ivectorPtr add_vector( ivectorPtr *root, char *name, char *vectorname, int vtype, char *dtype, int id )
{
ivectorPtr 	iv, v, p;
char 			*s;
int			camel;

if ( *name == 0 )return( NULL );

/*printf("Add vector [%s]/[%s] vtype %s dtype %s id %d\n", name, vectorname, vtype2string(vtype), dtype, id);*/

iv = calloc( 1, sizeof( ivector )); 
if ( iv == NULL ) { fprintf(stderr,"%%VC-F-NOMEM, couldn't allocate memory\n"); exit(4); }

iv->name   		= strdup( name  );
iv->vtype		= vtype;

if ( vtype == VALIAS && vectorname != NULL)iv->vectorname = strdup( vectorname );

iv->dtype		= strdup( dtype  );
for( s = iv->dtype; *s;++s) *s = toupper(*s);

iv->id  			= id;

/* go to the end of the list */
for ( p = *root, v = *root; v ; p = v, v = v->next );

if ( p != v ){
	p->next  	= iv;
}else{
	*root 		= iv;
}

return(iv);
}
/*------------------------------------------------------------------*/

void show_vector( ivectorPtr *root, int verbose )
{
ivectorPtr 	v;
int			count;


for ( v = *root, count=0; v ; v = v->next, ++count ){
	if ( verbose ){
		printf("%s%s%s\n    vectortype %s destinationtype %s\n    id %d written %d\n", 
               v->name, 
					v->vtype==VALIAS?"/":"",
					v->vtype==VALIAS?v->vectorname:"",
				   vtype2string( v->vtype ),
               v->dtype,
					v->id, v->written);
	}
}

if ( verbose ) fprintf(stderr,"%%VC-I-COUNT, Found %d vectors\n", count );


}
/*------------------------------------------------------------------*/

void free_vector( ivectorPtr *root )
{
ivectorPtr 	v,p;
int			count;

 	for ( v = *root; v ; ){
       free( v->name );
       free( v->vectorname );
       free( v->dtype );
       p = v;
       v = v->next;
       free(p);
	}
*root = NULL;
}

/*------------------------------------------------------------------*/

int file2vector( FILE *in, ivectorPtr *root , int id, int verbose)
{
char 	*filebuf, *s;
int 	len,maxlen,mode,namelen;
int   vtype;
int	retstat;
stat_t vecstat;
char  *v;
char	name[1024];
char  vectorname[1024];
char  typename[1024];
ivectorPtr tail=NULL;
int	bytesread,totalread;

retstat = fstat( fileno( in ), &vecstat );
if ( retstat ){
   fprintf(stderr,"%%VC-E-NOSTAT, Error stat of vector file\n");
   return( vaxc$errno );
}

filebuf = calloc( 1, (vecstat.st_size + 32) );
if ( filebuf == NULL ){
 	return( C$_ENOMEM );
}


for( bytesread = 0, totalread = 0;;){

      bytesread = decc$record_read(in, ( filebuf + totalread) , (vecstat.st_size - totalread) );

      if ( bytesread <= 0 ) break;
      if ( feof(in) ) break;

		totalread += bytesread;
		*(filebuf + totalread) = 0;
}

if ( verbose )fprintf( stderr, "%%VC-I-READ, read %d bytes from file %d\n", totalread,id);

name[0]  		= 0;
vectorname[0] 	= 0;
typename[0]  	= 0;

for ( mode = 0, s = filebuf; *s; ++s ){

	if ( *s == '!' || ( *s == '-' && isspace(*(s+1)) )  ){
       while( *s && *s != '\n' && *s != '\r' )++s;
		 --s;	
	    continue;		
	}

	 switch( mode ){
		case 0:
         if ( strncasecmp( s, "SYMBOL_VECTOR", 13) == 0 )mode = 1; 
			break;
      case 1:
			if (*s == '(' )mode=2;
			break;
		case 2: /* find start of vectorname */
			switch(*s){
				case ')':
					mode = 0;
					break;
				case '\r':
				case '\n':
				case '\t':
				case ' ':
				case ',':
				case '-':
					break;
				default:
               v = &name[0];
               namelen = 0;
					--s;
					mode = 3;
	            break;
			}
			break;
		case 3: /*fill vectornames */
			switch(*s){
				case ' ':
					break;
				case '=':	
         		*v = 0;
               mode = 4;
               v = &typename[0];
               namelen = 0;
					break;
				case '/':
               *v = 0;
					v = &vectorname[0];
               namelen = 0;
					break;
				default:
               if ( !(isspace(*s)) ){
						*v = *s;
						++v;
              	 	++namelen;
						if ( namelen > 128 ){
							fprintf(stderr, "%%VC-F-PARSEFAIL, failed parsing option file\n");
							exit(2);
						}
					}
				   break;
			}
			break;		
		case 4:
			if ( !(isspace(*s)) ){
          	--s;
				mode =5;
			}		
			break;
		case 5: /*fill type*/
			switch ( *s ){
				case ' ':
				case '\r':
				case '\n':
				case '\t':
				case ',':
				case ')':
               *v = 0;
               --s;
					mode = 6;
               break;
            default:
					*v = *s;
					++v;
					break;
			}
			break;
		case 6:

         vtype = VALIAS;
         if ( vectorname[0] == 0 )vtype = VVECTOR;
			tail = add_vector( tail==NULL?root:&tail, name, vectorname, vtype, typename, id );

			name[0] 			= 0;
			vectorname[0]  = 0;
			typename[0]  	= 0;

			mode = 0;
			if ( *s != ')' ) mode = 2;
			break;
		case 7:
			break;
 	}
}

free( filebuf );
return(1);
}

/*-----------------------------------------------------------------*/

int find_vector_diffs( ivectorPtr *sroot, ivectorPtr *croot )
{
ivectorPtr 	ov,cv;
int			status = 1;

for( ov = *sroot, cv = *croot; ov && cv; ov = ov->next, cv = cv->next ){
	if ( ov->vtype != cv->vtype || strcmp( ov->name, cv->name ) ){
      fprintf(stderr,"%%VC-I-DIFF, Found difference, old: %s, new: %s\n", ov->name, cv->name);
		status = 3;			
		break;
	}
}
if ( ov != cv ) status = 3;
return( status );
}

/*-----------------------------------------------------------------*/
/* code in compare_vector not necessarily optimized for performance 
 * but for intelligibility. 
 */

int compare_vector( ivectorPtr *sroot, ivectorPtr *nroot, ivectorPtr *croot, char *newfilename, int verbose )
{
ivectorPtr	ov, nv, cv, rv;
int			samename,samealiasname,namesameasalias, newversion;
char			newtypename[64];
char			newcomment[256];

for ( ov = *sroot; ov; ov = ov->next ){

   if ( ov->vtype == VVECTOR ){
      /* check for a one to one match */
		for ( nv = *nroot ; nv ; nv = nv->next){
          if ( nv->vtype == VVECTOR && nv->written == 0){
			 	if ( strcmp( ov->name , nv->name ) == 0 ){
					add_vector( croot, ov->name, NULL, ov->vtype, ov->dtype, 2 );
            	ov->written = 1;
            	nv->written = 1;
            	break;
			 	}
			 }		  	
		}
		/* there was no one to one match for this vector, look for a matching alias */
		if ( nv == NULL ){
		 for ( nv = *nroot ; nv ; nv = nv->next){
          if ( nv->vtype == VALIAS && nv->written == 0){
			 	if ( strcmp( ov->name , nv->name ) == 0 ){
					/* the old vector name matches a new alias, write an alias 
                * with the old name as an alias to the new vector.
			       */
					add_vector( croot, ov->name,nv->vectorname,nv->vtype, ov->dtype, 2 );
            	ov->written = 1;
            	nv->written = 1;
            	break;
			 	}
			 }		  	
		 }
		}

		/* look using an old alias; to solve this:
         old: een=procedure
              EEN/een=procedure
			new: EEN=procedure
              een_/EEN=procedure

			or this:
			old:	FOUR=procedure
					VIER/FOUR=procedure
			new:  vier=procedure
					VIER/vier=procedure	

		 */

		if ( nv == NULL ){

       for( cv = *sroot; cv; cv= cv->next ){
          if ( cv->vtype == VALIAS )if ( strcmp(cv->vectorname, ov->name) == 0 )break; 
		 }
       if ( cv != NULL ){
		 	for ( nv = *nroot ; nv ; nv = nv->next){
			 	if ( strcmp( cv->name , nv->name ) == 0 ){
					if ( nv->vtype == VVECTOR ){
					 add_vector( croot, ov->name,nv->name,VALIAS, ov->dtype, 2 );
					 fprintf(stderr,"%%VC-W-INDREFTYPO1, %s added as alias to %s, using indirect old alias %s/%s\n",
										ov->name, nv->name, cv->name, cv->vectorname);
            	 ov->written = 1;
					}
					if ( nv->vtype == VALIAS ){
					 add_vector( croot, ov->name,nv->vectorname,VALIAS, ov->dtype, 2 );
					 fprintf(stderr,"%%VC-W-INDREFTYPO2, %s added as alias to %s, using indirect old alias %s/%s\n",
										ov->name, nv->vectorname, cv->name, cv->vectorname);
            	 ov->written = 1;
					}
			 	}
            if ( ov->written) break;
		   }
		 } /* end if cv != NULL */	
		}



		if ( nv == NULL ){
	   	/* we give up, write entry as private_...*/
         if ( strncasecmp( "PRIVATE_", ov->dtype,8 ) ){
				sprintf( newtypename,"PRIVATE_%s", ov->dtype);
         }else{
				strcpy( newtypename,ov->dtype);
			}
			fprintf(stderr,"%%VC-W-REMOVED, %s vector %s removed, written as %s\n", ov->dtype, ov->name, newtypename);
			add_vector( croot, ov->name,"",VVECTOR,newtypename, 2 );
        	ov->written = 1;
		}
	}

   if ( ov->vtype == VALIAS ){

      /* check for a one to one match */
		for ( nv = *nroot ; nv ; nv = nv->next){
          if ( nv->vtype == VALIAS && nv->written == 0){
			 	if ( strcmp( ov->name , nv->name ) == 0 ){
					add_vector( croot, ov->name, nv->vectorname, ov->vtype, ov->dtype, 2 );
            	nv->written = 1;
            	ov->written = 1;
            	break;
			 	}
			 }		  	
		}

		/* there was no one to one match for this alias, look for a matching vector */
		if ( nv == NULL ){
		 for ( nv = *nroot ; nv ; nv = nv->next){
          if ( nv->vtype == VVECTOR ){
			 	if ( strcmp( ov->name , nv->name ) == 0 ){
					/* the old alias vector name matches a new vector, write a vector 
                * with the old name. 
			       */
					add_vector( croot, nv->name,NULL,nv->vtype, nv->dtype, 2 );
            	nv->written = 1;
            	ov->written = 1;
            	break;
			 	}
			 }		  	
		 }
		}

		/* there was no one to one match for this alias, look for a matching alias */
      /* solves this :
				old: twee_=PROCEDURE
					  TWEE/twee_=PROCEDURE
            new: twee_=PROCEDURE
					  TWEE_/twee_=PROCEDURE

       */

		if ( nv == NULL ){
		 for ( nv = *nroot ; nv ; nv = nv->next){
          if ( nv->vtype == VALIAS ){
			 	if ( strcmp( ov->vectorname , nv->name ) == 0 ){
					add_vector( croot, ov->name,nv->vectorname,ov->vtype, ov->dtype, 2 );
					fprintf(stderr,"%%VC-W-INDREFTYPO3, old alias %s/%s maintained as %s/%s\n",
										ov->name, ov->vectorname, ov->name, nv->vectorname);
            	ov->written = 1;
            	break;
			 	}
			 }		  	
          if ( nv->vtype == VVECTOR ){
			 	if ( strcmp( ov->vectorname , nv->name ) == 0 ){
					add_vector( croot, ov->name,nv->name,ov->vtype, ov->dtype, 2 );
					fprintf(stderr,"%%VC-W-INDREFTYPO4, old alias %s/%s maintained as %s/%s\n",
										ov->name, ov->vectorname, ov->name, nv->name);
            	ov->written = 1;
            	break;
			 	}
			 }		  	
		 }
		}

      /* give up */
		if ( nv == NULL ){
	   	/* we give up, write entry as private_...*/
         if ( strncasecmp( "PRIVATE_", ov->dtype,8 ) ){
				sprintf( newtypename,"PRIVATE_%s", ov->dtype);
         }else{
				strcpy( newtypename,ov->dtype);
			}
			fprintf(stderr,"%%VC-W-REMOVED, %s alias %s/%s removed, written as %s\n", ov->dtype, ov->name, ov->vectorname, newtypename);
			add_vector( croot, ov->name,"",VVECTOR,newtypename, 2 );
        	ov->written = 1;
		}
   } /* end if ov->vtype == VALIAS */
}

/* Add all new vectors to the end of croot list */


for ( newcomment[0] = 0, nv = *nroot; nv ; nv = nv->next ){

	  if ( nv->written == 0 ){

       if ( !newcomment[0] ){
		 	sprintf( newcomment,"! Newly added vectors and aliases %s %.200s",
					newfilename?"from":"", newfilename?newfilename:"");
		 	add_vector( croot, newcomment,NULL,VCOMMENT, "DATA", 2 );
       }

		 fprintf( stderr,"%%VC-I-ADD, %s %s %s%s%s added in new version\n", nv->dtype, 
				vtype2string(nv->vtype),nv->name,
            nv->vtype==VALIAS?"/":"",
				nv->vtype==VALIAS?nv->vectorname:"");

		 add_vector( croot, nv->name,nv->vectorname,nv->vtype, nv->dtype, 2 );
		 nv->written = 1;
	 }
}

return(1);
}

/*------------------------------------------------------------------*/

void vector2file( ivectorPtr *root, FILE *compatible , char format)
{
ivectorPtr 	v;
int			count;
char			continuestring[]=", -";
char			endstring[]=")";

fprintf( compatible, "CASE_SENSITIVE=YES\n");

if ( format == 'p' )fprintf( compatible, "SYMBOL_VECTOR=( -\n");

 	for ( v = *root, count=0; v ; v = v->next, ++count ){
       if ( v->vtype == VALIAS ){
        	if ( format == 'p')fprintf(compatible,"%s/%s=%s%s\n", 
											  v->name,v->vectorname,v->dtype,
											  (v->next==NULL)?endstring:continuestring );
		  if(  format == 'l')fprintf(compatible," symbol_vector=(%s/%s=%s)\n", 
											  v->name,v->vectorname, v->dtype);
		 }
       if ( v->vtype == VVECTOR ){
        	if ( format == 'p')fprintf(compatible,"%s=%s%s\n", 
											  v->name,v->dtype,
											  (v->next==NULL)?endstring:continuestring );
		   if(  format == 'l')fprintf(compatible," symbol_vector=(%s=%s)\n", 
											  v->name, v->dtype);
		 }
       if ( v->vtype == VCOMMENT ){
        	 fprintf(compatible,"%s\n",v->name);
		 }

	}

fprintf( compatible, "CASE_SENSITIVE=NO\n");

return;
}

/*------------------------------------------------------------------*/
void close_files( FILE *e, FILE *t, FILE *d)
{
if (e) fclose(e);
if (t) fclose(t);
if (d) fclose(d);

}

/*------------------------------------------------------------------*/
int print_help()
{
fprintf(stderr,"Usage : vectorcompatible -s <oldvectorfile> -n <newervectorfile> -o <outputvectorfile> [ -f [p|l] ]\n");
fprintf(stderr," Take an (-s) old linker options file with symbol vectors describing\n");
fprintf(stderr," an older version of a shareable library. Take a linker options\n");
fprintf(stderr," file (-n) with symbol vectors of a newer version of this library.\n");
fprintf(stderr," Vectorcompatible writes a symbol options file (-o), that can be used\n");
fprintf(stderr," to produce an upward compatible version of the shareable library.\n");
fprintf(stderr," -f format of symbol vector file, format p or l are allowed:\n");
fprintf(stderr,"    p: symbol_vector=(a=procedure,b=procedure,...)\n");
fprintf(stderr,"    l: symbol_vector=(a=procedure)\n symbol_vector=(b=procedure)\n ...\n");
fprintf(stderr,"    default is format 'p'.\n");
fprintf(stderr," -v verbose.\n");
 
return( SS$_BADPARAM - 4 );
}
/*------------------------------------------------------------------*/

main(int argc, char **argv)
{
FILE			*standard=NULL,*newnot=NULL,*compatible=NULL;
int			namearg[3];
ivectorPtr	svector=NULL,nvector=NULL,cvector=NULL;
int			i,status, aliasisupcase=1, verbose=0;
time_t		nowtime;
char			format='p';

if ( argc < 6 ){
   return( print_help() );
}

for( i = 1; i < argc; ++i ){
	switch( argv[i][0] ){
		case '-':
    		switch( argv[i][1] ){
               case 's':
						++i;
						if ( i >= argc ){fprintf( stderr,"%%VC-E-NONAME, Missing standard vector filename\n");return( SS$_BADPARAM - 2);}
						standard = fopen( argv[i],"r");
						if ( ! standard ){
   						fprintf( stderr, "%%VC-E-OPENERR, Couldn't read old standard vectorfile %s\n", argv[i]);
   						close_files(standard,newnot,compatible);
   						return( vaxc$errno );
						}
						namearg[0] = i;
						break;
					case 'n':
						++i;
						if ( i >= argc ){fprintf( stderr,"%%VC-E-NONAME, Missing new incompatible vector filename\n");return( SS$_BADPARAM - 2);}
						newnot = fopen( argv[i],"r");
						if ( ! newnot ){
   						fprintf( stderr, "%%VC-E-OPENERR, Couldn't read new incompatible vectorfile %s\n", argv[i]);
   						close_files(standard,newnot,compatible);
   						return( vaxc$errno );
						}
						namearg[1] = i;
						break;
               case 'o':
						++i;
						if ( i >= argc ){fprintf( stderr,"%%VC-E-NONAME, Missing new compatible vectorfilename\n");return( SS$_BADPARAM - 2);}
						compatible = fopen( argv[i],"w");
						if ( ! compatible ){
						   fprintf( stderr, "%%VC-E-OPENERR, Cannot write new compatible vectorfile %s\n", argv[3]);
						   close_files(standard,newnot,compatible);
						   return( vaxc$errno );
						}
						namearg[2] = i;
						break;
					case 'f':				
						++i;
						if ( i >= argc ){fprintf( stderr,"%%VC-E-ILLFORMAT, Missing format argument\n");return( SS$_BADPARAM - 2);}
						if ( tolower(argv[i][0]) == 'p' || tolower(argv[i][0]) == 'l' ){
                   	format = tolower( argv[i][0] );
						}else{
                   	fprintf(stderr,"%%VC-W-ILLFORMAT, Illegal format %c, will use format p\n", tolower( argv[i][0] ) );
						}
						break;
					case 'h':
                  return( print_help() );
						break;
					case 'v':
                  verbose = 1;
						break;

         }
			break;
		default:
				fprintf( stderr,"%%VC-W-IGNPARAM, Ignoring parameter %s\n", argv[i] );
				break;
	}
}



status = file2vector( standard, &svector , 0, verbose);
if ( !(status&1) )return( status );


 if ( verbose ) fprintf(stderr,"Show Vectors in standard vector file\n");
 show_vector( &svector,verbose );


status = file2vector( newnot, &nvector , 1, verbose);
if ( !(status&1) )return( status );

/* if ( verbose ) fprintf(stderr,"Show Vectors in new vector file\n");
   show_vector( &nvector,verbose ); 
 */

compare_vector( &svector, &nvector, &cvector, argv[ namearg[1] ], verbose );

/* if ( verbose ) fprintf(stderr,"Show Vectors in compatible vector file\n");
 * show_vector( &cvector,verbose ); 
 */

status = find_vector_diffs( &svector, &cvector );
 
nowtime = time(0);
fprintf( compatible,"! generated by vectorcompatible.exe %s", asctime(localtime( &nowtime )) );
fprintf( compatible,"! based on %.120s\n! and %.126s\n", argv[ namearg[0] ], argv[ namearg[1] ]);

vector2file( &cvector, compatible, format );

close_files(standard,newnot,compatible);



free_vector( &nvector );
free_vector( &svector );
free_vector( &cvector );

exit( status );
}

$ on error then goto errorcompile
$ cc vectorcompatible.c
$ link vectorcompatible.obj
$ if f$search("vectorcompatible.*;*") .nes. "" then delete vectorcompatible.*;*/exclude=vectorcompatible.exe
$!
$ vectorcompatible :== $'f$environment("default")'vectorcompatible.exe
$!
$exit 1
$Errorcompile:
$!
$ say "Error compiling vectorcompatible.c, abort"
$ exit 4 
$endsubroutine
$!
$!--------------------------------------------------------------------
$!
$MakeListLib: subroutine
$!
$ set noon
$ if f$type( listlib ) .nes. "" 
$ then
$      define/user sys$output nl:
$      define/user sys$error  nl:
$		 listlib
$      if f$message($status,"ident") .eqs. "%BADPARAM" then exit 1     
$ endif  
$!
$ say "Compile and link listlib"
$ type/out=interpretmap.h sys$input:
/* in interpretmap.c */
int read_versionmap( char *mapname, char *version, char *excludecsvlist );
void free_versionmaps();
int is_globalsym(char *version, char *candidate);

$ type/out=interpretmap.c sys$input:
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stat.h>

#include <ssdef.h>
#include <errno.h>
#include <errnodef.h>
#include <unixlib.h>

#define MAPLOCAL 	0;
#define MAPGLOBAL 1;

typedef struct ssymtab{
	char 		*camelname;
   struct   ssymtab *next;
}symtab;

typedef struct versiondependency{
	char					*version;
   char					*parentversion;
	symtab				*globals;
	symtab				*locals;
	int					parentcount;
	struct versiondependency *parent;
	struct versiondependency *next;
}versionmap;

static versionmap	*versionlist;
static versionmap	*excludelist;

/*---------------------------------------------------*/
/* For every version in the version script, a list   */

versionmap *add_to_versionlist( char *thisversion )
{
versionmap  *l,*p;

l = calloc( 1, sizeof( versionmap ) );

if ( versionlist == NULL ){
	versionlist = l;
}else{	
	for ( p = versionlist; p->next ;p = p->next );
   p->next = l;
}

l->version  = strdup( thisversion );

return(l);
}

/*---------------------------------------------------*/
/* add symbolname or wildcard to a local or a global 
 * symbol list of a versionmap
 */

symtab *add_to_symlist( symtab **symlist, char *camelname )
{
symtab  *l,*p;

l = calloc( 1, sizeof( symtab ) );

if ( *symlist == NULL ){
	*symlist = l;
}else{	
	for ( p = *symlist; p->next ;p = p->next );
   p->next = l;
}

l->camelname  = strdup( camelname );

return(l);
}

/*---------------------------------------------------*/
void print_symlist( symtab *symlist)                   
{
symtab  *l;

for ( l = symlist; l ; l = l->next )printf("  %s\n", l->camelname);

}

/*---------------------------------------------------*/

int is_globalsym(char *version, char *candidate)
{
versionmap *m,*h;
int		  highparent=-1;
symtab	  *l;
	
/* if version is empty take the version with most
   parents: should be the newest version.
 */
if ( versionlist == NULL ) return( 7);
	
for ( m = versionlist, h = NULL; m; m = m->next){
	 if ( version != NULL && *version ){	
	 	 if ( strcmp( version, m->version) == 0 ){
          h = m;
			 break;
		 }		
    }else{
		 if ( m->parentcount > highparent ){
          h = m;
			 highparent = m->parentcount;	
		 }	
	 }
}

for ( m = h ; m ; m = m->parent ){

	if ( m->globals ) {
		for ( l = m->globals; l ; l = l->next ){
       	if( decc$match_wild( candidate, l->camelname ) )return(1);
		}
	}
	if ( m->locals ) {
		for ( l = m->locals; l ; l = l->next ){
       	if( decc$match_wild( candidate, l->camelname ) )return(0);
		}
	}
}

if ( excludelist != NULL ){
		for ( l = excludelist->locals; l ; l = l->next ){
       	if( decc$match_wild( candidate, l->camelname ) )return(2);
		}
		/* if exclude list is given, default is global */
      return(3);
}

/* no match found. return 5 (global)*/

return(5);
}
/*---------------------------------------------------*/

print_versionsymbols(char *version)
{
versionmap *m,*h;
int		  highparent=-1;

/* if version is empty take the version with most
   parents: should be the newest version.
 */
	
for ( m = versionlist; m; m = m->next){
	 if ( version != NULL ){	
	 	 if ( strcmp( version, m->version) == 0 ){
          h = m;
			 break;
		 }		
    }else{
		 if ( m->parentcount > highparent ){
          h = m;
			 highparent = m->parentcount;	
		 }	
	 }
}

for ( m = h ; m ; m = m->parent ){
	if ( m->parentcount == -1 ) continue;
	printf("version: %s, %d parents%s%s\n", m->version, m->parentcount,
											m->parentversion?", depends on ":"",
										   m->parentversion?m->parentversion:"" );
	if ( m->globals ) {
		printf(" globals:\n");
      print_symlist( m->globals );
	}
	if ( m->locals ) {
		printf(" locals:\n");
      print_symlist( m->locals );
	}
}
if ( excludelist != NULL ){
		printf(" locals in excludelist:\n");
      print_symlist( excludelist->locals );
}

}
/*---------------------------------------------------*/
void free_symlist( symtab **symlist)
{
symtab  *l, *p;

for ( l = *symlist; l;){
		free( l->camelname );
      p = l->next;
		free( l );
		l = p;		
}
*symlist = NULL;

}

/*---------------------------------------------------*/

void free_versionmaps()
{
versionmap  *l, *p;


for ( l = versionlist; l;){
		free( l->version );
		free( l->parentversion );
		free_symlist( &(l->globals));
		free_symlist( &(l->locals));
      l->parent = NULL;
      p = l->next;
		free( l );
		l = p;		
}
versionlist = NULL;

}

/*---------------------------------------------------*/
/* Have every versionmap point to it's parent, and
   update the parentcount of each versionmap.
 */

int update_versionmap( versionmap *vroot )
{
versionmap *l,*p;
char		  *parentname;

/* fill parent pointers */	
for ( l = vroot; l ; l = l->next ){
   if ( l -> parentversion != NULL ){
 	 for ( p = vroot; p ; p = p->next ){
   	 if ( strcmp( l->parentversion, p->version ) == 0 ){
          l->parent = p;
			 break;	
		 }	
	 }
	}
}
/*update parentcount */
for ( l = vroot; l ; l = l->next ){
   if ( l -> parentversion != NULL ){
 	 for ( p = l->parent; p ; p = p->parent ){
    	l->parentcount += 1;
    }
	}
//   fprintf(stderr, "Found %d parents for %s\n", l->parentcount, l->version);

}
return(1);
}

/*-----------------------------------------------------------*/
/* in exclude list a number of (wildcarded) symbols that should
 * remain local.
 */

int add_exclude_syms( char *excludecsvlist )
{
char			*thisex,*e,*mylist;
int			mode;

if ( excludecsvlist == NULL ) return(3);
if ( *excludecsvlist == 0 ) return(3);

mylist = strdup( excludecsvlist );
 
excludelist = add_to_versionlist( "excludelist" );
excludelist->parentcount = -1;

for( mode = 0, e = mylist; mode > -1 ; ++e ){ 
	switch( mode ){
		case 0:
			if ( !(isspace( *e )) && *e != ',' ){
				thisex = e;
				mode = 1;
			}
			break;
		case 1:
			if ( isspace( *e ) || *e == ',' || *e == 0 ){
				mode = 0;				
            if ( *e == 0 ) mode = -1;
				*e = 0;
				add_to_symlist( &(excludelist->locals), thisex );					
			}			
			break;
	}
}		

free( mylist );
return(1);
}
/*---------------------------------------------------*/
int get_version_symbols( char *maps, char *version )
{
char	*l,*s,*foundsymbol;
int	mode, symbolcount;
int	globloc;
char	*thisversion, *dependversion, *thisdepend;
versionmap *thisversionmap;
symtab	  **thissymtab;

if ( maps == NULL ) return(0);

for( mode = 0, symbolcount = 0, l = maps; *l ; ++l ){
//   printf("mode %d, %c (%d)\n", mode, *l, *l);

	if ( *l == '#' ){
       while( *l && *l != '\n' && *l != '\r' )++l;
	    while( *l && isspace(*l) )++l;
		 --l;	
	    continue;		
	}
//	if ( *l == '\n' || *l == '\r' )continue;

	switch( mode ){
		case 0: /* find version */
             if ( ! isspace(*l) ){
					 thisversion = l;
					 mode = 1;
				 }
				 break;
		case 1: /* read past version */
				if ( isspace(*l) ){
					*l = 0;
				}					
            	if ( *l == '{' ){
					*l = 0;
					mode = 2;
               		thisversionmap = add_to_versionlist( thisversion );
               		thissymtab = &(thisversionmap->globals);
				}
				break;
		case 2: /* get first symbol */
				if ( *l == '}' ){
             	mode = 4;
					break;
				}
				if ( ! isspace(*l) && *l != ';' && *l != '{' ){
               foundsymbol = l;
					mode = 3;
					break;
            }
				break;
		case 3: /* read to end of symbol */
            if ( *l == ':' ){
               if ( *(l - 6) == 'g' ){
                  thissymtab = &(thisversionmap->globals);
               }else{
               	thissymtab = &(thisversionmap->locals);
					}
               foundsymbol = NULL;
					mode = 2;
				}
				if ( isspace(*l) || *l == ';' ){
					*l = 0;

					add_to_symlist( thissymtab, foundsymbol );					

               foundsymbol = NULL;
					++symbolcount;
					mode = 2;
				}
      		break;
      case 4: /*find version this version depends on*/
				if ( !(isspace(*l)) ){
					if ( *l != ';' ){
						thisdepend = l;
						mode = 5;
					}else{
						/*no dependencies */
                  mode = 0;
					}
				}
				break;
		case 5:
			  if (  isspace( *l) || *l == ';' ){
              	*l = 0;
					thisversionmap->parentversion = strdup( thisdepend );
					mode = 0;
			  }		
			  break;	
   }
}

if ( mode == 0 ){
	return( symbolcount );
}
return(-1);
}
/*---------------------------------------------------*/
/* Read a UNIX type version script, to determine what 
   symbols are global and should be exported, and what
   symbols are local, and should remain unmentioned
 */

int read_versionmap( char *mapname, char *version, char *excludecsvlist )
{
FILE	*mapfile;
char	*mapbuffer;
stat_t mapstat;
int 	 retstat, bytesread, totalread;
int	 nocomment;
int	 symbolsfound;

if ( excludecsvlist ){
   if ( excludecsvlist[0] != 0)add_exclude_syms( excludecsvlist );
}
if ( mapname == NULL ){
   return(1);
}
if ( mapname[0] == 0 )return(1);

mapfile = fopen( mapname, "r" );
if ( mapfile == NULL ){
   fprintf(stderr,"%%LL-E-OPENMAP, Error opening version map file %s\n", mapname);
   return( vaxc$errno );
}

retstat = fstat( fileno( mapfile ), &mapstat );
if ( retstat ){
   fprintf(stderr,"%%LL-E-NOSTAT, Error stat of version map file %s\n", mapname);
   fclose( mapfile );
   return( vaxc$errno );
}

mapbuffer = calloc( 1, (mapstat.st_size + 32) );
if ( mapbuffer == NULL ){
 	return( C$_ENOMEM );
}

for( bytesread = 0, totalread = 0;;){

        bytesread = decc$record_read(mapfile, (mapbuffer + totalread) , (mapstat.st_size - totalread) );

        if ( bytesread <= 0 ) break;
        if ( feof(mapfile) ) break;

		printf("read %d bytes\n",bytesread);

		totalread += bytesread;
		*(mapbuffer + totalread) = 0;
}


symbolsfound = get_version_symbols( mapbuffer, version ); 
if ( symbolsfound > 0 ){
	retstat = 1;
	update_versionmap(versionlist);

}else{
   retstat = 0;
	fprintf(stderr, "  Error interpreting symbols in %s\n", mapname);
}


free( mapbuffer );
fclose( mapfile );

return( retstat );
}

#ifdef INTERPRETMAP_MAIN
/*-----------------------------------------------------------*/

main(int argc, char **argv)
{
int	status;
char	*version=NULL;
char	*excludecsvlist=NULL;
char	*symbolname=NULL;
char	*mapname=NULL;

if ( argc > 5 ){
	fprintf(stderr, "Usage: interpretmap <versionmapfile> <version> <symbolname> <excludelist>\n");
	return( SS$_INVARG );
}	

if ( argc > 1 && argv[1][0] ) mapname = argv[1];
if ( argc > 2 && argv[2][0] ) version = argv[2];
if ( argc > 3 && argv[3][0] ) symbolname = argv[3];
if ( argc > 4 && argv[4][0] ) excludecsvlist = argv[4];

status = read_versionmap( mapname, version, excludecsvlist );
if ( (status&1) ){
	if ( symbolname == NULL ){
		print_versionsymbols( version );
 	}else{
		printf("%s is %s\n", symbolname,
						  ((is_globalsym( version, symbolname ))&1)?"a global symbol":"a LOCAL symbol");
	}
}else{
 	fprintf(stderr,"error processing %s\n", mapname );
}
free_versionmaps(); 

}

#endif

$ type/out=listlib.c sys$input:
/* the shortening routines are provided under the following copyright: */
/*
 * Copyright (c) 2010 Craig A. Berry
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
/* The rest of the code is provided under the same conditions, but the copyright
 * is Jose Baars. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#define __NEW_STARLET 1

#include <starlet.h>
#include <ssdef.h>
#include <iosbdef.h>
#include <efndef.h>
#include <iledef.h>
#include <syidef.h>
#include <descrip.h>
#include <lbr$routines.h>
#include <lib$routines.h>
#include <lbrdef.h>
#include <str$routines.h>

#include "interpretmap.h"

#define UINT32 unsigned int



struct libentry{
char              shortname[256];
char              name[256];
char              shortupname[256];
char              type[32];
struct libentry   *next;
};

struct libentry *globroot;

/*----------------------------------------*/
/* update architecture to 1 for VAX, 
 * 2 for Alpha and 3 for IA64 
 */

int get_architecture(int *architecture)
{
IOSB  iob;
ILE3  syi_items[2];
int   status,index;

index = 0;

syi_items[index].ile3$w_length       = 4;
syi_items[index].ile3$w_code         = SYI$_ARCH_TYPE;
syi_items[index].ile3$ps_bufaddr     = architecture;
syi_items[index].ile3$ps_retlen_addr = NULL;

++index;

syi_items[index].ile3$w_length       = 0;
syi_items[index].ile3$w_code         = 0;
syi_items[index].ile3$ps_bufaddr     = NULL;
syi_items[index].ile3$ps_retlen_addr = NULL;
 


status = sys$getsyiw( EFN$C_ENF, 0,0,&syi_items[0], &iob, 0,0);
if ( !(status&1) ) return( status);
if ( !(iob.iosb$w_status&1))return( iob.iosb$w_status );

return(status);
}

/*----------------------------------------*/

UINT32 crc32( const char *inputdata )
{
  UINT32 crc32;
  UINT32 seed = ~0UL;
  UINT32 coefficient = 0x0EDB88320UL; /* AUTODIN II */
  UINT32 crcarray[16];

  struct  dsc$descriptor inputdatad;

  inputdatad.dsc$w_length   = strlen( inputdata );
  inputdatad.dsc$b_dtype    = DSC$K_DTYPE_T;
  inputdatad.dsc$b_class    = DSC$K_CLASS_S;
  inputdatad.dsc$a_pointer  = (char *)inputdata;

  lib$crc_table( (void *) &coefficient, (void *) crcarray );
  crc32 = lib$crc( (void *) crcarray, (void *) &seed, &inputdatad );

  return crc32;
}

/*----------------------------------------*/
/*
 * This is the RFC2938 variant of base32, not RFC3548, Crockford's, or
 * other newer variant.  It produces an 8-byte encoded character string
 * (plus trailing null) from a 32-bit integer input.
 */

void u32_to_base32(UINT32 input, char *output)
{
    static const char base32hex_table[32] = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
        'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
        'u', 'v'
    };
    int i;

    /*
     * Grab lowest 5 bits and look up conversion in table.  Lather, rinse,
     * repeat for 6, 5-bit chunks to accommodate 32 bits of input.
     */
    for (i = 0; i < 7; i++) {
        output[6 - i] = base32hex_table[input & 0x1f];
        input >>= 5;     /* position to look at next 5 */
    }
    output[7] = '$';     /* It's DEC, so use '$' not '=' to pad. */
    output[8] = '\0';
}

/*----------------------------------------*/

void shorten_name_upcase( struct libentry *le )
{
char           b32str[9];
UINT32   crc;
char           *c, *u, upname[256];


    for( c = le->name, u = upname; *c; ++c, ++u ){
         *u = toupper( *c );
    }
    *u = 0; 

    if ( ( u - upname ) < 32 ){
       strcpy( le->shortupname, upname );
       return;
    }

    crc = crc32( upname );

    (void) u32_to_base32(crc, (char *)&b32str);

    for (c = b32str; *c; c++){
         *c = toupper(*c);
    }

    sprintf( le->shortupname, "%.23s%.8s\n", upname, b32str);
    le->shortupname[31] = '\0';
}

/*----------------------------------------*/
struct libentry *addlibentry( struct libentry **libroot, char *shortname, char *name )
{
struct libentry *newle, *le, *pe;

newle = calloc( 1, sizeof( struct libentry ));
if ( newle == NULL ){
   fprintf(stderr,"%%LL-E-NOMEM, Error allocating memory for library entry\n");
   return(NULL);
}

strcpy( newle->shortname, shortname );
strcpy( newle->name,      name );

for ( le = *libroot, pe = *libroot; le; le = le->next)pe = le;

if ( pe == NULL ){
   *libroot = newle;
}else{
   pe->next  = newle;
}

return( newle );
}

/*----------------------------------------*/

void  showlibentries( struct libentry **libroot )
{
struct libentry *le;

for ( le = *libroot; le; le = le->next){
    printf("- shortname: %s\n       name: %s\n   shupname: %s\n",le->shortname, le->name, le->shortupname);
}

}
/*----------------------------------------*/

void  freelibentries( struct libentry **libroot )
{
struct libentry *le,*pe;

for ( le = *libroot, pe = *libroot; le; ){
   pe = le;
   le = le->next;    
   free( pe );
}

*libroot = NULL;

}

/*----------------------------------------*/

int   add_lbrentry(  struct dsc$descriptor *named, long long *rfa )
{
int      namelen;
struct   libentry *newle, *le, *pe;
char     *u,*s;

for ( le = globroot, pe = globroot; le; le = le->next){
   pe = le;
	named->dsc$a_pointer[ named->dsc$w_length ] = 0;
   if ( strcmp( le->shortname, named->dsc$a_pointer ) == 0 )return(1);
}

 
newle = calloc( 1, sizeof( struct libentry ));
if ( newle == NULL ){
   fprintf(stderr,"%%LL-E-NOMEM, Error allocating memory for library entry\n");
   return(0);
}

namelen = (int) named->dsc$w_length;

/* As these names have not been found in the demagler output,
   add them as is, just upcase the name in shortupname. 
*/

strncpy( newle->name, named->dsc$a_pointer, namelen);
newle->name[ namelen ] = 0;

strcpy( newle->shortname, newle->name );

shorten_name_upcase( newle );    

if ( pe == NULL ){
   globroot = newle;
}else{
   pe->next  = newle;
}

return(1);
}


/*----------------------------------------*/
int listlib( char *libname, struct libentry **libroot )
{
int                     status;
struct dsc$descriptor   libnamed;
unsigned int            libindex,libfunction,indexnumber;


libnamed.dsc$w_length  = strlen( libname );
libnamed.dsc$b_dtype   = DSC$K_DTYPE_T;
libnamed.dsc$b_class   = DSC$K_CLASS_S;
libnamed.dsc$a_pointer = libname;

libfunction            = LBR$C_READ;

status = lbr$ini_control( &libindex, &libfunction,0,0);
if ( !(status&1) ){
   fprintf( stderr, "%%LL-E-LBRERR, lbr$ini_control error\n");
   return( status );
}
 
status = lbr$open( &libindex, &libnamed,0,0,0,0,0);
if ( !(status&1) ){
   fprintf(stderr, "%%LL-E-LBROPENERR, lbr$open error\n");
   return( status );
}

/* in object libraries, global symbol names are in index 2 */

indexnumber = 2;

/* lbr$getindex doesn't offer the opportunity to
   pass on a user defined address.
*/

status = lbr$get_index( &libindex, &indexnumber, add_lbrentry, 0,0);
if ( !(status&1) ){
   fprintf(stderr,"%%LL-E-LBRIDXERR, libr$get_index error\n");
   status = lbr$close( &libindex );
   return( status );
}

*libroot    = globroot;

status = lbr$close( &libindex );
return( status );
}


/*----------------------------------------*/
int read_demangler( struct libentry **libroot, char *filename )
{
FILE     *d;
int      num;
char     line[1024]; 
char     name[256],shortname[256];
struct   libentry *le;  

d = fopen( filename,"r");
if ( d == NULL ){
   fprintf(stderr,"%%LL-E-OPENDEM, Error opening demangler file %s\n", filename);
   return( vaxc$errno );
}

while(1){
        num = decc$record_read(d, &line[0], 728);
        if ( num < 0) break;
        if ( feof(d) ) break;

        line[num] = 0;
        --num;

        for(;num; --num){
            if ( !isspace( line[num] ) )break;
            line[num] = 0; 
        }   
         
        sprintf( shortname, "%.31s", line );    
        sprintf( name,      "%s", &line[31]);      
         
        le = addlibentry( libroot, shortname, name );
        if ( le == NULL ){
             fclose( d );
             return( 2 );
        }   
      

        shorten_name_upcase( le );     
}

fclose( d );

globroot = *libroot;

return (1);
}

/*----------------------------------------*/
/* Object analysis parsing of both Alpha ans IA64 can extract
   both data- and function global sysmbols, but only data
   symbols are extracted. The library listing is supported
   and guaranteed to work, the object analysis extract is not
   that reliable.
*/

int read_alpha_objanalysis( char *analyzename, struct libentry **dataroot )
{
FILE  *ana=NULL;
int   foundeggs;
char  *eggstart;
char  regel[136];
char  *symname;
int   mode;

ana = fopen( analyzename,"r");
if ( !ana ){
   fprintf(stderr,"%%LL-E-OPENERR, Couldn't open ana/obj output file %s for read\n", analyzename);
   return( vaxc$errno);
}

/*  mode = -1 : verify it's an ana/obj file, ok if EOBJ$C_EGSD is found
    mode =  0 : look for EGSY$V_DEF == 1
    mode =  1 : look for EGSY$V_NORM == 1 or 0
    mode =  2 : look for 'symbol:'

*/
mode      = -1;
foundeggs = 0;

while ( fgets(regel, 132, ana) != NULL ){


      switch ( mode ){
       case -1:
            if ( strstr( regel, "(EOBJ$C_EGSD)") != NULL ){
               mode=0;
            }
            break;
       case 0:
            foundeggs = 0;
            eggstart = strstr( regel, "EGSY$V_DEF");
            if ( eggstart != NULL ){
               if ( strstr( eggstart, "1") != NULL ){
                     foundeggs |= 1;
                     mode=1;
               }
            }
            break;
       case 1:
            eggstart = strstr( regel, "EGSY$V_NORM" );
            if ( eggstart != NULL ){
               if ( strstr( eggstart, "1") != NULL ){
                     foundeggs |= 2;
                     mode=2;
               }
               if ( strstr( eggstart, "0") != NULL ){
                     foundeggs |= 4;
                     mode=2;
               }
            }
            break;
       case 2:
            eggstart = strstr( regel, "symbol:" );
            if ( eggstart != NULL ){
               if ( foundeggs == 0 ){
                  mode = 0;
                  break;
               }
               eggstart = strstr( eggstart, "\"" );
               ++eggstart; /*read past quotes */
               symname = eggstart;
               eggstart = strstr( eggstart + 1, "\"" );
               *eggstart = 0;

               if ( (foundeggs&4) ){
                   addlibentry( dataroot, symname, symname);
                   fprintf( stderr,"%%LL-I-FNDDATA, data symbol found: %s\n", symname);
               }

               mode = 0;
           }
           break;
         default:
           fprintf(stderr,"%%LL-E-PARSERR, Parse error of Alpha object analysis.\n");
           break;
      }/*end switch*/
}/*end while*/

fclose( ana );
return(1);
}
/*----------------------------------------*/

int read_ia64_objanalysis( char *analyzename, struct libentry **dataroot )
{
FILE    *objfile;
char    regel[1024];
char    *sstart, *ssend;
char    symname[4096];
int     i,mode, symdex;


objfile  = fopen( analyzename,"r");
if( objfile == NULL ) {fprintf( stderr,"%%LL-E-OPENERR, Unable to open obj analysis %s\n", analyzename); return( vaxc$errno ); }

mode = 0;
while( fgets( regel, 1022, objfile ) ){
   if( ( sstart = strstr( regel, "Symbol")) != NULL ){
         switch( mode ){
           case 0:
            if ( isdigit( *(sstart + 8) ) ){
               sstart = strchr( sstart,'\"') + 1;
               if ( *sstart != '\"'){
                  ssend = strchr( sstart,'\"');
                  sprintf( symname, "%.*s", ssend - sstart, sstart);
                  mode = 1;
               }
            }
            break;
           case 1:
            if ( strstr( sstart,"Symbol Type:") != NULL ){
               symdex = -1;
               if ( strstr( sstart, "STT_FUNC") != NULL ) symdex = 0;
               if ( strstr( sstart, "STT_OBJECT") != NULL ) symdex = 1;
               switch( symdex ){
                  case 0: /* function */
                     mode = 2;
                     break;
                  case 1: /* data */
                     mode = 3;
                     break;
                  default:
                     mode = 0;
               }
            }
            break;
        } /*end switch */
   } /* end if symbol */

   if ( mode == 2 ){
      if ( ( sstart = strstr( regel,"Linkage Type")) != NULL ){
         if ( strstr( sstart, "VMS_STL_STD") != NULL ){
            mode = 3;
         }else{
            mode = 0;
         }
      }
    }

   if ( mode == 3 ){
      if ( ( sstart = strstr( regel,"Bound to section:")) != NULL ){
         if ( strstr( sstart, "$CODE$") != NULL ||
              strstr( sstart, symname)  != NULL ){
               if ( symdex == 1 ){
                  // add to data symbol list
                   addlibentry( dataroot, symname, symname);
                   fprintf( stderr,"%%LL-I-FNDDATA, data symbol found: %s\n", symname);
               }
         }
         mode = 0;
     }
   }
} /* end while */

fclose( objfile );
return(1);
}

/*----------------------------------------*/

int isdatasymbol( struct libentry *g, struct libentry *dataroot)
{
struct libentry *d;

for ( d = dataroot; d; d = d->next ){
   if ( strcmp( d->shortname, g->shortname) == 0 || 
        strcmp( d->shortname, g->shortupname) == 0){
        return( 1 );
   }
}
return(0);
}
/*----------------------------------------*/
int write_ofile( FILE *ofile, struct libentry *g, struct libentry *dataroot,char format, int first)
{

/* find symbol type */
if ( isdatasymbol( g, dataroot) ){
   sprintf( g->type,"DATA");
}else{
   sprintf( g->type,"PROCEDURE");
}

/* write an alias ? */
if ( strcmp( g->shortname, g->shortupname ) ){
   if ( format == 'p' ){
      fprintf( ofile,"%s -\n%s=%s, -\n%s/%s=%s", first?"":",",
         g->shortname, g->type, g->shortupname,g->shortname, g->type);
   }
   if ( format == 'l' ){
      fprintf( ofile,"symbol_vector=(%s=%s)\nsymbol_vector=(%s/%s=%s)\n", 
         g->shortname, g->type, g->shortupname,g->shortname, g->type);
   }
}else{
   if ( format == 'p' )fprintf( ofile,"%s -\n%s=%s", first?"":",",g->shortname,g->type);
   if ( format == 'l' )fprintf( ofile,"symbol_vector=(%s=%s)\n",g->shortname, g->type);
}


return(1);
}
/*----------------------------------------*/

int write_vector( struct libentry *dataroot, char *mapname, char *mapversion, char *excludelist, char *outputname, char format)
{
FILE  *ofile;
struct libentry *g;
int				 status, first=1;
time_t 			 nowtime;

status = read_versionmap( mapname, mapversion, excludelist );
if ( !(status&1) ){
	fprintf(stderr,"%%LL-E-ILLMAP, error processing %s\n", mapname );
   return( status );
}

ofile = fopen( outputname, "w");
if( ofile == NULL ){fprintf(stderr,"%%LL-E-OPENOUT, Error opening output file %s\n", outputname); return( vaxc$errno); }

nowtime = time(0);
fprintf( ofile,"! generated by listlib.exe %s", asctime(localtime( &nowtime )) );
fprintf( ofile,"CASE_SENSITIVE=YES\n");
if ( format == 'p' )fprintf( ofile,"SYMBOL_VECTOR=(");


for ( g = globroot; g ; g = g->next ){
	 if ( ((is_globalsym( mapversion, g->name ))&1) ){
         write_ofile( ofile, g, dataroot, format, first);
			first = 0;
    }
}

if ( format == 'p' )fprintf( ofile,")\n");
fprintf( ofile,"CASE_SENSITIVE=NO\n");

fclose ( ofile );

free_versionmaps(); 
return(1);
}

/*----------------------------------------*/
int print_help()
{
   fprintf( stderr,"Usage: listlib <object library>\n");
   fprintf( stderr,"      Writes a linker vector options file with global symbol names as found in an object library\n");       
   fprintf( stderr,"      If uppercase aliases are needed, the objects must be compiled with the AS_IS switch.\n");       
   fprintf( stderr,"      -d demangler_db filename\n");
   fprintf( stderr,"       Based on the cxx$demanglerdb produced by the C-compiler when used with flag SHORTENED,\n");       
   fprintf( stderr,"       it will produce both upper and lowercase shortened symbols, if necessary.\n");       
   fprintf( stderr,"       The mangling as done by the C++ compiler works differently and is not supported.\n");       
   fprintf( stderr,"      -a <analyze/obj output of objectlibrary>\n");
   fprintf( stderr,"       For the obj analysis(used to determine the type, PROCEDURE or DATA, of the\n");
   fprintf( stderr,"       global symbol, first extract all objects like this:\n");
   fprintf( stderr,"       $ library/extract=*/out=objects.t <objectlibrary>\n");
   fprintf( stderr,"       Then analyze/object output should be generated like this:\n");       
   fprintf( stderr,"       On Alpha: $  analyze/object/gsd 'objects'.t  /out=objanalysis.t\n");       
   fprintf( stderr,"       On IA64 : $  analyze/object/sect=symtab 'objects'.t /out=objanalysis.t\n");
   fprintf( stderr,"      -e a quoted, comma separated list with possibly wildcarded symbols\n");
   fprintf( stderr,"       that must be excluded form the output (commonly symbols starting with _).\n");
   fprintf( stderr,"       the list is case-sensitive, names are compared to the not-shortened name of the symbol\n");
   fprintf( stderr,"      -o write to this file in stead of SYS$OUTPUT.\n");
   fprintf( stderr,"      -m Unix type version script (usually called xxx.map) filename\n");
   fprintf( stderr,"         all symbols not mentioned in this version script will be global\n");
   fprintf( stderr,"      -n Version to use for the version script, literal version as used in script.\n");
   fprintf( stderr,"         If no version is supplied use version with most dependencies.\n");
   fprintf( stderr,"      -f format of symbol vector file, format p or l are allowed:\n");
   fprintf( stderr,"        p: symbol_vector=(a=PROCEDURE,b=PROCEDURE,...)\n");
   fprintf( stderr,"        l: symbol_vector=(a=PROCEDURE)\n           symbol_vector=(b=PROCEDURE)\n            ...\n");
   fprintf( stderr,"       default is format 'p'.\n");
   fprintf( stderr,"      Example\n");
   fprintf( stderr,"      $ library/extract=*/out=libz.objects libz.olb\n");
   fprintf( stderr,"      $ pipe ana/obj/gsd libz.objects | listlib libz.olb -a sys$input: -e \"_*\" -o libz_vec.opt\n");
   fprintf( stderr,"      produces a linker options file libz_vec.opt, without symbols starting with an underscore\n");
   return( SS$_BADPARAM - 4 );
}

/*----------------------------------------*/

main( int argc, char **argv )
{
char              libname[4096], manglename[4096],analyzename[4096],outputname[4096];
char              mapname[4096], mapversion[4096];
char              excludelist[4096];
char              format[32];
int               i,status;
struct libentry   *libroot=NULL;
struct libentry   *mangleroot=NULL;
struct libentry   *dataroot=NULL;
int               architecture;

libname[0]     = 0;
manglename[0]  = 0;
analyzename[0] = 0;
outputname[0]  = 0;
excludelist[0] = 0;
format[0]      = 0;
mapname[0]     = 0;
mapversion[0]  = 0;

for( i = 1; i < argc; ++i ){
   switch( argv[i][0] ){
      case '-':
         switch( argv[i][1]){
            case 'a':
                  ++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NOOBJANA, Missing object analysis output name\n");return(2);}
                  strcpy( analyzename, argv[i]);
                  break;
            case 'd':
                  ++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NODEMAN, Missing demangler db name\n");return(2);}
                  strcpy( manglename, argv[i]);
                  break;
            case 'e':
                  ++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NOEXCL, Missing excludelist\n");return(2);}
                  strcpy( excludelist, argv[i]);
                  break;
            case 'f':
                  ++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NOVECFORM, Missing format specifier\n");return(2);}
                  strcpy( format, argv[i]);
                  break;
            case 'm':
						++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NOVERSIONSCRIPT, Missing version script name\n");return(2);}
                  strcpy( mapname, argv[i]);
                  break;
            case 'n':
						++i;
                  if ( i >= argc ){fprintf( stderr,"%%LL-E-NOVERSION, Missing version\n");return(2);}
                  strcpy( mapversion, argv[i]);
                  break;
            case 'h':
                  return( print_help() );
                  break;
            case 'o':
                  ++i;
                  if ( i >= argc ){fprintf(stderr,"%%LL-E-NOOUT, Missing outputfilename\n");return(2);}
                  strcpy( outputname, argv[i]);
                  break;
            default:
                  fprintf(stderr, "%%LL-W-IGNSWI, Ignoring unknown switch -%c\n", argv[i][1]);
                  break;
         }
         break;   
      default:
         if ( libname[0] == 0 ){
            strcpy( libname, argv[i] );
         }else{
            fprintf(stderr,"%%LL-W-IGNPAR, Ignoring parameter [%s]\n", argv[i]);
         }
         break;
   }
}

if ( libname[0] == 0 ){
   return( print_help() );
}

globroot    = NULL;

if ( outputname[0] == 0 ){
   strcpy(outputname, "SYS$OUTPUT:");
}

if ( format[0] ){
   format[0] = tolower( format[0] );
   if ( format[0] != 'p' && format[0] != 'l' ){
      fprintf(stderr,"%%LL-W-UNVECFORM, Unknown format %c, default of 'p' will be used\n", format[0]);
      format[0] = 'p';
   }
}else{
      format[0] = 'p';
}

if ( analyzename[0] ){
   status = get_architecture( &architecture );
   /* create a list with all DATA (i.e. non-PROCEDURE ) symbols */
   status = 0;
   /* it is possible to read an IA64 object analysis on Alpha and vice versa,
    * but that would require a lot of explaining on what to do with the outcome.
    */
   if ( architecture == 1 ){ status = SS$_UNSUPPORTED; fprintf(stderr,"%%LL-E-NOVAX, VAX is not supported\n"); }
   if ( architecture == 2 )status = read_alpha_objanalysis( analyzename, &dataroot );
   if ( architecture == 3 )status = read_ia64_objanalysis( analyzename, &dataroot );
   if ( !(status&1) ){ fprintf(stderr,"%%LL-E-ERROBJANA, Reading object analysis failed, status %d\n", status);return( status);}
}

/* printf("reading demanger?\n");
 */

if ( manglename[0] ){
   status = read_demangler( &mangleroot, manglename );
   if ( !(status&1) ) return( status);
}else{
   fprintf(stderr, "%%LL-W-NODEMANG, No demangler db used. If functions are shortened, vectors are incorrect.\n");
}

status = listlib( libname, &libroot );
if ( !(status&1) ) return( status);

/* printf("Show all objectlib symbols\n");
 * showlibentries(&globroot );
 * printf("Show all data symbols\n");
 * showlibentries(&dataroot );
 */

status = write_vector( dataroot, mapname, mapversion, excludelist, outputname, format[0] );


freelibentries(&mangleroot );
freelibentries(&dataroot );
freelibentries(&globroot );

return( status );

}

$ on error then goto errorcompile
$ cc interpretmap.c
$ cc listlib.c
$ link/exe=listlib.exe listlib.obj, interpretmap.obj
$ if f$search("interpretmap.*;*") .nes. "" then delete interpretmap.*;*
$ if f$search("listlib.*;*") .nes. "" then delete listlib.*;*/exclude=listlib.exe
$!
$ listlib :== $'f$environment("default")'listlib.exe
$!
$exit 1
$Errorcompile:
$!
$ say "Error compiling listlib.c, abort"
$ exit 4 
$!
$endsubroutine
